#include "walls.h"
#include <QtXml/qdom.h>
#include <QFile>
#include "../common/sources/system/fileCache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "game.h"

CWalls   walls;

CWalls::CWalls()
{

}

// For each tile, the file name and coordinates of the graphics for the "back", "left" and "right" walls.
struct CDispTile
{
    char        file[16];
    uint16_t    x, y;
};

// Tiles are sorted from left to right horizontally and "rows" of tiles are sorted from the farthest to the nearest

static const CDispTile   gDispTiles[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH][3] =
{
    // Row 3 (farthest)
    {
        {{"", 0, 0}, {"", 0, 0},               {"", 0, 0}               }, // 03
        {{"", 0, 0}, {"Wall13_L.png", 8, 58},  {"", 0, 0}               }, // 13
        {{"", 0, 0}, {"Wall23_L.png", 78, 59}, {"Wall23_R.png", 134, 59}}, // 23
        {{"", 0, 0}, {"", 0, 0},               {"Wall33_R.png", 180, 58}}, // 33
        {{"", 0, 0}, {"", 0, 0},               {"", 0, 0}               }  // 43
    },

    // Row 2
    {
        {{"Wall02_F.png",   0, 58}, {"", 0, 0},               {"", 0, 0}               }, // 02
        {{"Wall12_F.png",   7, 58}, {"Wall12_L.png",  0, 57}, {"", 0, 0}               }, // 12
        {{"Wall22_F.png",  77, 58}, {"Wall22_L.png", 60, 52}, {"Wall22_R.png", 146, 52}}, // 22
        {{"Wall32_F.png", 146, 58}, {"", 0, 0},               {"Wall32_R.png", 216, 57}}, // 32
        {{"Wall42_F.png", 216, 58}, {"", 0, 0},               {"", 0, 0}               }  // 42
    },

    // Row 1
    {
        {{"", 0, 0},                {"", 0, 0},               {"", 0, 0}               }, // 01
        {{"Wall11_F.png",   0, 52}, {"", 0, 0},               {"", 0, 0}               }, // 11
        {{"Wall21_F.png",  59, 52}, {"Wall21_L.png", 33, 42}, {"Wall21_R.png", 164, 42}}, // 21
        {{"Wall31_F.png", 164, 52}, {"", 0, 0},               {"", 0, 0}               }, // 31
        {{"", 0, 0},                {"", 0, 0},               {"", 0, 0}               }  // 41
    },

    // Row 0 (nearest)
    {
        {{"", 0, 0},                {"", 0, 0},              {"", 0, 0}               }, // 00
        {{"Wall10_F.png",   0, 42}, {"", 0, 0},              {"", 0, 0}               }, // 10
        {{"Wall20_F.png",  32, 42}, {"Wall20_L.png", 0, 33}, {"Wall20_R.png", 191, 33}}, // 20
        {{"Wall30_F.png", 191, 42}, {"", 0, 0},              {"", 0, 0}               }, // 30
        {{"", 0, 0},                {"", 0, 0},              {"", 0, 0}               }  // 40
    }
};

//--------------------------------------------------------------------------------------------------------
// return the sequence in which we will parse the table
CVec2       CWalls::getDrawSequence(CVec2 pos)
{
    CVec2   tablePos;

    // in x: because of the perspective and because some walls overlaps, we need to draw the
    // leftmost/rightmost tiles first, then the center ones.
    static const int xSequence[] = {0, 4, 1, 3, 2};
    tablePos.x = xSequence[pos.x];

    // in "depth": we draw the farthest tiles first then the nearest ones.
    // That's already the sequence of the table.
    tablePos.y = pos.y;

    return tablePos;
}

//--------------------------------------------------------------------------------------------------------
// returns the map position from the position in the table
// this position is relative to the player and we assume we are looking toward the north (top of the map)

CVec2       CWalls::getMapPos(CVec2 tablePos)
{
    return CVec2(tablePos.x - 2,
                 tablePos.y - WALL_TABLE_HEIGHT + 1);
}

//--------------------------------------------------------------------------------------------------------
// returns the table index for the given wall direction
int       CWalls::getTableSide(EWallSide side, bool flip)
{
    if (side == eWallSideUp)
        return 0;

    if (side == eWallSideLeft)
    {
        if (flip == false)
            return 1;
        else
            return 2;
    }

    if (side == eWallSideRight)
    {
        if (flip == false)
            return 2;
        else
            return 1;
    }

    return 0;
}

//--------------------------------------------------------------------------------------------------------
const char* CWalls::getFilename(CVec2 tablePos, EWallSide side, bool flip)
{
    int sideIndex = getTableSide(side, flip);
    const char* baseName;

    if (flip == false)
        baseName = gDispTiles[tablePos.y][tablePos.x][sideIndex].file;
    else
        baseName = gDispTiles[tablePos.y][WALL_TABLE_WIDTH - 1 - tablePos.x][sideIndex].file;

    if (baseName[0] != 0)
    {
        static char fileName[256];
        sprintf(fileName, "gfx/3DView/walls/%s", baseName);
        return fileName;
    }
    return "";
}

//--------------------------------------------------------------------------------------------------------
CVec2       CWalls::getGfxPos(CVec2 tablePos, EWallSide side, bool flip)
{
    int sideIndex = getTableSide(side, flip);
    const CDispTile*  dispTile;

    if (flip == false)
        dispTile = &gDispTiles[tablePos.y][tablePos.x][sideIndex];
    else
        dispTile = &gDispTiles[tablePos.y][WALL_TABLE_WIDTH - 1 - tablePos.x][sideIndex];

    int x = dispTile->x;

    if (flip == true)
    {
        QImage  image = fileCache.getImage(getFilename(tablePos, side, flip));
        x = MAINVIEW_WIDTH - x - image.size().width();
    }

    return CVec2(x, dispTile->y);
}

//--------------------------------------------------------------------------------------------------------
void    CWalls::readOrnatesDB()
{
    QDomDocument doc;

    QFile f("databases/ornates.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement ornate = root.firstChild().toElement();

    while(!ornate.isNull())
    {
        if (ornate.tagName() == "ornate")
        {
            COrnateData   newOrnate;

            newOrnate.name = ornate.attribute("name");

            QDomElement ornateInfo = ornate.firstChild().toElement();

            while(!ornateInfo.isNull())
            {
                if (ornateInfo.tagName() == "image_front")
                {
                    newOrnate.imageFront = QString("gfx/3DView/ornates/") + ornateInfo.text();
                }
                else if (ornateInfo.tagName() == "image_side")
                {
                    newOrnate.imageSide = QString("gfx/3DView/ornates/") + ornateInfo.text();
                }
                else if (ornateInfo.tagName() == "pos_front")
                {
                    newOrnate.posFront.x = ornateInfo.attribute("x").toInt();
                    newOrnate.posFront.y = ornateInfo.attribute("y").toInt();
                }
                else if (ornateInfo.tagName() == "pos_side")
                {
                    newOrnate.posSide.x = ornateInfo.attribute("x").toInt();
                    newOrnate.posSide.y = ornateInfo.attribute("y").toInt();
                }

                ornateInfo = ornateInfo.nextSibling().toElement();
            }
            ornatesDatas.push_back(newOrnate);
        }
        ornate = ornate.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CWalls::readWallsDB()
{
    QDomDocument doc;

    QFile f("databases/walls.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement wall = root.firstChild().toElement();

    while(!wall.isNull())
    {
        std::vector<CParamType> paramsTypesList;

        if (wall.tagName() == "wall")
        {
            QDomElement wallInfo = wall.firstChild().toElement();

            while(!wallInfo.isNull())
            {
                if (wallInfo.tagName() == "param")
                {
                    QString paramType = wallInfo.attribute("type");
                    CParamType  newParamType;

                    if (paramType == "ornate")
                    {
                        newParamType.mType = eParamOrnate;
                        newParamType.mName = wallInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "int")
                    {
                        newParamType.mType = eParamInt;
                        newParamType.mName = wallInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "bool")
                    {
                        newParamType.mType = eParamBool;
                        newParamType.mName = wallInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "enum")
                    {
                        newParamType.mType = eParamEnum;
                        newParamType.mName = wallInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "stack")
                    {
                        newParamType.mType = eParamStack;
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "string")
                    {
                        newParamType.mType = eParamString;
                        newParamType.mName = wallInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                }

                wallInfo = wallInfo.nextSibling().toElement();
            }
            map.mWallsParams.push_back(paramsTypesList);
        }
        wall = wall.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
// table with the pos and size of each walls for the ornates
struct SWallInfos
{
    CVec2   pos;
    CVec2   size;
};

static const SWallInfos   gWallsInfos[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH][3] =
{
    // Row 3 (farthest)
    {
        {{{0, 0}, {0, 0}}, {{0, 0}, {0, 0}},     {{0, 0}, {0, 0}}     }, // 03
        {{{0, 0}, {0, 0}}, {{ 8, 58}, {36, 48}}, {{0, 0}, {0, 0}}     }, // 13
        {{{0, 0}, {0, 0}}, {{78, 59}, {12, 48}}, {{134, 59}, {12, 48}}}, // 23
        {{{0, 0}, {0, 0}}, {{0, 0}, {0, 0}},     {{180, 58}, {36, 48}}}, // 33
        {{{0, 0}, {0, 0}}, {{0, 0}, {0, 0}},     {{0, 0}, {0, 0}}     }  // 43
    },

    // Row 2
    {
        {{{-77, 58}, {70, 48}}, {{0, 0}, {0, 0}},     {{0, 0}, {0, 0}}     }, // 02
        {{{  7, 58}, {70, 48}}, {{0, 0}, {0, 0}},     {{0, 0}, {0, 0}}     }, // 12
        {{{ 77, 58}, {70, 48}}, {{60, 52}, {18, 74}}, {{146, 52}, {18, 74}}}, // 22
        {{{146, 58}, {70, 48}}, {{0, 0}, {0, 0}},     {{0, 0}, {0, 0}}     }, // 32
        {{{216, 58}, {70, 48}}, {{0, 0}, {0, 0}},     {{0, 0}, {0, 0}}     }  // 42
    },

    // Row 1
    {
        {{{0, 0}, {0, 0}},       {{0, 0}, {0, 0}},      {{0, 0}, {0, 0}}      }, // 01
        {{{-46, 52}, {105, 74}}, {{0, 0}, {0, 0}},      {{0, 0}, {0, 0}}      }, // 11
        {{{ 59, 52}, {105, 74}}, {{33, 42}, {27, 111}}, {{164, 42}, {27, 111}}}, // 21
        {{{164, 52}, {105, 74}}, {{0, 0}, {0, 0}},      {{0, 0}, {0, 0}}      }, // 31
        {{{0, 0}, {0, 0}},       {{0, 0}, {0, 0}},      {{0, 0}, {0, 0}}      }  // 41
    },

    // Row 0 (nearest)
    {
        {{{0, 0}, {0, 0}},         {{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}, // 00
        {{{-127, 42}, {159, 111}}, {{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}, // 10
        {{{  32, 42}, {159, 111}}, {{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}, // 20
        {{{ 191, 42}, {159, 111}}, {{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}, // 30
        {{{0, 0}, {0, 0}},         {{0, 0}, {0, 0}}, {{0, 0}, {0, 0}}}  // 40
    }
};

//---------------------------------------------------------------------------------------------
CRect CWalls::drawOrnate(QImage* image, CVec2 tablePos, EWallSide side, int ornate, int textLines)
{
    if (ornate != 0)
    {
        int tableSide = walls.getTableSide(side, false);

        if (gWallsInfos[tablePos.y][tablePos.x][tableSide].size.x != 0)
        {
            COrnateData& ornateData = walls.ornatesDatas[ornate];
            const SWallInfos*   ornateTabData = &gWallsInfos[tablePos.y][tablePos.x][tableSide];
            const SWallInfos*   ornate0 = NULL;
            QString ornateFileName;
            CVec2   ornatePos;
            bool    ornateFlip = false;

            switch(side)
            {
                case eWallSideUp:
                    ornate0 = &gWallsInfos[3][2][tableSide];
                    ornateFileName = ornateData.imageFront;
                    ornatePos = ornateData.posFront;
                break;

                case eWallSideLeft:
                    ornate0 = &gWallsInfos[2][2][1];
                    ornateFileName = walls.ornatesDatas[ornate].imageSide;
                    ornatePos = ornateData.posSide;
                break;

                case eWallSideRight:
                    ornate0 = &gWallsInfos[2][2][1];
                    ornateFileName = walls.ornatesDatas[ornate].imageSide;
                    ornatePos = ornateData.posSide;
                    ornateFlip = true;
                break;

                default:
                    break;
            }

            // calculate the position and scale factor and draw the ornate in a temporary image
            float   scale = (float)ornateTabData->size.y / 111.0f;
            QImage  ornateImage = fileCache.getImage(ornateFileName.toLocal8Bit().constData());

            if (ornate == ORNATE_TEXT)
            {
                CRect   rect;
                rect.tl = CVec2(0, 0);
                rect.br.x = ornateImage.width() - 1;

                if (side == eWallSideUp)
                {
                    static int heights[] = {0, 14, 27, 41, 55};
                    rect.br.y = heights[textLines];
                }
                else
                {
                    static int heights[] = {0, 10, 20, 31, 41};
                    rect.br.y = heights[textLines];
                }
                ornateImage = graph2D.cropImage(ornateImage, rect);
            }
            QSize   ornateSize = ornateImage.size() * scale;
            ornatePos = (ornatePos - ornate0->pos);
            ornatePos.x = ornatePos.x * ornateTabData->size.x / ornate0->size.x;
            ornatePos.y = ornatePos.y * ornateTabData->size.y / ornate0->size.y;

            if (ornateFlip == true)
                ornatePos.x = ornateTabData->size.x - ornatePos.x - ornateSize.width();

            ornatePos += ornateTabData->pos;

            QImage  tempImage(ornateSize, QImage::Format_ARGB32);
            tempImage.fill(QColor(0, 0, 0, 0));
            graph2D.drawImageScaled(&tempImage, CVec2(), ornateImage, scale, ornateFlip);

            // darken the ornate based on it's distance
            static const float shadowLevels[] = {0.0f, 0.2f, 0.4f};
            float shadow;

            if (side == eWallSideUp)
                shadow = shadowLevels[WALL_TABLE_HEIGHT - 1 - tablePos.y];
            else
                shadow = shadowLevels[WALL_TABLE_HEIGHT - 2 - tablePos.y];

            graph2D.darken(&tempImage, shadow);

            // draw the ornate on screen.
            graph2D.drawImage(image, ornatePos, tempImage, 0, false, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));

            // return the ornate rect
            return CRect(ornatePos, ornatePos + CVec2(ornateSize.width() - 1, ornateSize.height() - 1));
        }
    }
    return CRect();
}

//---------------------------------------------------------------------------------------------
void CWalls::init()
{
    // create stacks for wall torches
    for (int y = 0; y < map.mSize.y; ++y)
        for (int x = 0; x < map.mSize.x; ++x)
        {
            CVec2   pos(x, y);
            CTile*  t = map.getTile(pos);

            for (int side = 0; side < eWallSideMax; ++side)
            {
                CWall* w = &t->mWalls[side];

                if (w->getType() == eWallTorch)
                {
                    bool    isEmpty = w->getBoolParam("isEmpty");

                    map.removeObjectsStack(pos, (EWallSide)side);

                    if (isEmpty == false)
                    {
                        CObjectStack*   stack = map.addObjectsStack(pos, (EWallSide)side);
                        CObject object;
                        object.setType(3);
                        stack->addObject(object);
                    }
                }
            }
        }
}
